#include <stdio.h>
#include <string.h>
#pragma hdrstop

/*
   Coolpix RAW to Coolpix NEF converter
   by paul69, v0.04, 2003-01-26, <e2500@narod.ru>

   Reading EXIF JPEG file
*/

#include "tiff.h"
#include "raw2nef.h"

////////////////////////////////////////////////////////////////////////////////

struct TIFF_ID
{
   unsigned short tag;
   unsigned short type;
   unsigned count;
   unsigned offset;
};

class TObjTIFF;

typedef bool (*TParser)(TObjTIFF&,TIFF_ID&);
unsigned get_intel_dword(TObjTIFF& tiff);

class TObjTIFF
{
public:
   TObjTIFF(char* _buf, unsigned _size, unsigned _offset) :
      m_buf(_buf), m_size((int)_size), m_offset(_offset), m_start(0),
      binary_format(unknown_format)
      {}

   bool init();

   char* m_buf;
   int m_size;
   int m_i;
   unsigned m_offset;
   unsigned m_start;

   unsigned (*_get_word)(TObjTIFF&);
   unsigned (*_get_dword)(TObjTIFF&);
   unsigned (*_get_value)(TIFF_ID const&);

   void seek(unsigned _i) { m_i=_i-m_start; }
   unsigned tell() const { return m_i+m_start; }
   void skip(unsigned _i) { m_i+=_i; }

   void set_buff(char* _buf, unsigned _size, unsigned _offset )
   {
      m_buf = _buf;
      m_size = _size;
      m_offset = _offset;
      m_start = _offset;
      m_i = 0;
   }

   void check(int);
   unsigned char* get_ptr(int sz) { check(sz); return (unsigned char*)m_buf+m_i; }
   unsigned char get_byte() { check(1); return (unsigned char)m_buf[m_i++]; }
   unsigned get_word() { check(2); return _get_word(*this); }
   unsigned get_dword() { check(4); return _get_dword(*this); }
   unsigned get_dword_intel() { check(4); return get_intel_dword(*this); }
//   unsigned get_value(TIFF_ID const& tid) { return _get_value(tid); }


   void parse_tags(TParser parser, IFDir* );

   enum { unknown_format, format_intel, format_motorolla } binary_format;

//   void find_pars(TExifData&);

//   static int sizeof_value[n_types];
//   static const char* nameof_value[n_types];
//   static const char* GetTypeName( unsigned type );
};

class TBuffer
{
public:
   TBuffer(unsigned sz) : size(sz), buff(new char[sz]) {}
   ~TBuffer() { delete [] buff; }

   operator char* () { return buff; }
   char& operator [] (int i) { return buff[i]; }

   char* buff;
   unsigned size;
};


unsigned get_intel_dword(TObjTIFF& tiff)
{
   unsigned dw = *(unsigned*)(tiff.m_buf+tiff.m_i);
   tiff.m_i += 4;
   return dw;
}
unsigned get_intel_word(TObjTIFF& tiff)
{
   unsigned dw = *(unsigned short*)(tiff.m_buf+tiff.m_i);
   tiff.m_i += 2;
   return dw;
}

unsigned get_motorolla_dword(TObjTIFF& tiff)
{
   union _dw {
      unsigned char b[4];
      unsigned dw;
   } v;
   v.b[3] = tiff.m_buf[tiff.m_i]; ++tiff.m_i;
   v.b[2] = tiff.m_buf[tiff.m_i]; ++tiff.m_i;
   v.b[1] = tiff.m_buf[tiff.m_i]; ++tiff.m_i;
   v.b[0] = tiff.m_buf[tiff.m_i]; ++tiff.m_i;
   return v.dw;
}

unsigned get_motorolla_word(TObjTIFF& tiff)
{
   union _w {
      unsigned char b[2];
      unsigned short w;
   } v;

   v.b[1] = tiff.m_buf[tiff.m_i]; ++tiff.m_i;
   v.b[0] = tiff.m_buf[tiff.m_i]; ++tiff.m_i;
   return v.w;
}

////////////////////////////////////////////////////////////////////////////////

bool TObjTIFF::init()
{
   if( memcmp( m_buf, "II*\0", 4) == 0 )
   {
      binary_format = format_intel;
      _get_dword = get_intel_dword;
      _get_word = get_intel_word;
//      _get_value = get_intel_value;
   }
   if( memcmp( m_buf, "MM\0*", 4) == 0 )
   {
      binary_format = format_motorolla;
      _get_dword = get_motorolla_dword;
      _get_word = get_motorolla_word;
//      _get_value = get_motorolla_value;
   }

   if( binary_format == unknown_format ) { printf("bad TIFF header"); return false; }

   m_i = 4;

   return true;
}

void TObjTIFF::check(int sz)
{
   try
   {
      if( m_i+sz > m_size )
         throw "bound error";
   }
   catch(...)
   {
   }
}

////////////////////////////////////////////////////////////////////////////////
void add_tag(IFDir* IFD, TObjTIFF& tiff, TIFF_ID const& tid)
{
   unsigned k;
   IFDir::Tag* tag = IFD->add( tid.tag, tid.type, tid.count, NULL );

   if( !tag ) return;
   
   switch( tid.type )
   {
   case TIFF::type_UNDEFINED:
   case TIFF::type_BYTE:
   case TIFF::type_SBYTE:
   case TIFF::type_ASCII:

      for(k=0; k<tid.count; ++k) {
         ((byte*)tag->value)[k] = tiff.get_byte();
      } break;

   case TIFF::type_WORD:
   case TIFF::type_SHORT:

      for(k=0; k<tid.count; ++k) {
         ((word*)tag->value)[k] = (word)tiff.get_word();
      } break;

   case TIFF::type_DWORD:
   case TIFF::type_LONG:

      for(k=0; k<tid.count; ++k) {
         ((dword*)tag->value)[k] = tiff.get_dword();
      } break;

   case TIFF::type_RATIONAL:
   case TIFF::type_SRATIONAL:

      for(k=0; k<2*tid.count; ++k) {
         ((dword*)tag->value)[k] = tiff.get_dword();
      } break;

   default:
      break;
   }
}

////////////////////////////////////////////////////////////////////////////////


bool is_NIKON;
bool ParseNikonMakerTag(TObjTIFF& tiff, TIFF_ID& tid );

IFDir* pIFD;
IFDir* pExif;
IFDir* pMakerNote;

bool ParseTag( TObjTIFF& tiff, TIFF_ID& tid )
{
   switch( tid.tag )
   {
   case 0x10E: return true; // ImageDescription
   case 0x10F:
      if( memcmp(tiff.get_ptr(tid.count), "NIKON", 5) == 0 ) is_NIKON = true;
      return true;
   case 0x110: return true; // Model always E5700
   case 0x112: return true; // Orientation
   case 0x11A: return true; // XResolution
   case 0x11B: return true; // YResolution
   case 0x128: return true; // ResolutionUnit

   case 0x131: return true; // Software
   case 0x132: return true; // DateTime
   case 0x13E: return true; // WhitePoint

   case 0x214: return true; // ReferenceBlackWhite

   case 0x927C:
   {
      unsigned temp = tiff.tell();
      if( memcmp(tiff.get_ptr(6), "Nikon\0", 6) == 0 )
      {
         tiff.skip( 6 );
         int sub_type = tiff.get_dword_intel();
         if( sub_type == 2 )
         {
            // Nikon TIFF IFD
            int size = tid.count - 10;
            if( size > tiff.m_size-tiff.tell() )
               size = tiff.m_size-tiff.tell();

            TObjTIFF tiff2(
               (char* )tiff.get_ptr( size ),
               size,
               tiff.m_offset+tiff.tell() );

            if( tiff2.init() )
            {
               // IFD0
               unsigned ifd = tiff2.get_dword();
               if( ifd )
               {
                  tiff2.seek(ifd);
                  tiff2.parse_tags(ParseNikonMakerTag, pMakerNote);
               }
               tiff.seek( temp );
               return true;
            }
         }
      }
      else
      if( is_NIKON )
      {
         // Nikon IFD
         tiff.seek( tid.offset );
         tiff.parse_tags(ParseNikonMakerTag, pMakerNote);
         tiff.seek( temp );
         return true;
      }

      return true;
   }

   case 0x8769: return false; // Exif IFD
   case 0x8825: return false; // GPS IFD
   case 0xA005: return false; // Interoperability IFD

   case 0x828D: return false; // CFARepeatPatternDim
   case 0x828E: return false; // CFAPattern

   case 0x8298: return true; // Copyright
   case 0x829A: return true; // ExposureTime
   case 0x829D: return true; // F-Number

   case 0x8773: return true; // ICC profile

   case 0x8822: return true; // Exposure Program
   case 0x8824: return true; // SpectralSensitivity
   case 0x8827: return true; // ISO
   case 0x8828: return true; // OptElectricConvFactor

   case 0x9000: return false; // ExifVersion
   case 0x9003: return true; // DateTimeOriginal
   case 0x9004: return true; // DateTimeDigitized

   case 0x9101: return false; // ComponentsConfiguration
   case 0x9102: return false; // CompressedBitsPerPixel

   case 0x9201: return true; // ShutterSpeedValue
   case 0x9202: return true; // Aperture Value
   case 0x9203: return true; // BrightnessValue
   case 0x9204: return true; // Exposure Bias
   case 0x9205: return true; // MaxApertureValue
   case 0x9206: return true; // SubjectDistance
   case 0x9207: return true; // Metering Mode
   case 0x9208: return true; // Light Source
   case 0x9209: return true; // Flash
   case 0x920A: return true; // FocalLength
   case 0x9214: return true; // SubjectArea

   case 0x9286: return true; // UserComment
   case 0x9290: return true; // SubSecTime
   case 0x9291: return true; // SubSecTimeOriginal
   case 0x9292: return true; // SubSecTimeDigitized

   case 0xA000: return false; // FlashPixVersion
   case 0xA001: return true; // ColorSpace
   case 0xA002: return false; // PixelXDimension for compressed only
   case 0xA003: return false; // PixelYDimension for compressed only
   case 0xA004: return true; // RelatedSoundFile

   case 0xA20B: return true; // FlashEnergy
   case 0xA20C: return true; // SpatialFrequencyResponse
   case 0xA20E: return true; // FocalPlaneXResolution
   case 0xA20F: return true; // FocalPlaneYResolution
   case 0xA210: return true; // FocalPlaneResolutionUnit
   case 0xA214: return true; // SubjectLocation
   case 0xA215: return true; // ExposureIndex
   case 0xA217: return true; // SensingMethod
   case 0xA300: return true; // FileSource
   case 0xA301: return true; // SceneType
   case 0xA302: return false; // ColorFilterArrayPattern

   case 0xA401: return true; // CustomRendered
   case 0xA402: return true; // ExposureMode
   case 0xA403: return true; // WhiteBalance
   case 0xA404: return true; // DigitalZoomRatio
   case 0xA405: return true; // FocalLengthIn35mmFilm
   case 0xA406: return true; // SceneCaptureType
   case 0xA407: return true; // GainControl
   case 0xA408: return true; // Contrast
   case 0xA409: return true; // Saturation
   case 0xA40A: return true; // Sharpness
   case 0xA40B: return true; // DeviceSettingDescription
   case 0xA40C: return true; // SubjectDistanceRange
   default: break;
   }

   return false;
}

// ExifReader 2.80 (http://www.takenet.or.jp/~ryuuji/minisoft/exifread/english/)
bool ParseNikonMakerTag(TObjTIFF& tiff, TIFF_ID& tid )
{
   switch( tid.tag )
   {
   case 0x001: return false; // Version
   case 0x002: return true; // ISO Setting
   case 0x003: return true; // ColorMode
   case 0x004: return false; // Quality always RAW
   case 0x005: return true; // WhiteBalance
   case 0x006: return true; // Sharpening
   case 0x007: return true; // AutoFocus Mode
   case 0x008: return true; // Flash Sync Mode
   case 0x00A: return true; // ImageDiagonal
   case 0x00F: return true; // ISO Selection
   case 0x010: return false; // Dump
   case 0x011: return false; // thumbnail
   case 0x080: return true; // ImageAdjustment //   
   case 0x081: return true; // Image Adjustment // ExifReader 2.80
   case 0x082: return true; // AuxiliaryLens
   case 0x084: return true; // Lens
   case 0x085: return true; // Manual Focus
   case 0x086: return true; // Digital Zoom
   case 0x088: return true; // AF Position
   case 0x089: return true; // ShotMode // ExifReader 2.80
   case 0x08D: return true; // ColorSpace // ExifReader 2.80
   case 0x08F: return true; // Scene
   case 0x092: return true; // Hue Adjustment // ExifReader 2.80
   case 0x094: return true; // Saturation // ExifReader 2.80
   case 0x095: return true; // Noise Reduction
   case 0x100: return false;
   case 0xE00: return false; // PrintIM

   default: return ParseTag( tiff, tid );
   }
}

////////////////////////////////////////////////////////////////////////////////


void TObjTIFF::parse_tags(TParser parser, IFDir* IFD)
{
   try
   {
      unsigned number_of_fields = get_word();

      unsigned Exif_IFD = 0;
      unsigned temp, temp2;

      TIFF_ID tid;
      unsigned i;
      for(i=0; i < number_of_fields; ++i, seek( temp ) )
      {
         tid.tag = (unsigned short)get_word();
         tid.type = (unsigned short)get_word();
         tid.count = get_dword();

         unsigned data_pos = tell();
         tid.offset = get_dword();

         if( tid.type >= TIFF::n_types ) continue;
         
         unsigned size = TIFF::sizeof_value[tid.type]*tid.count;

         // if size of data less 4 bytes, then offset hold data
         if( tid.type >= TIFF::n_types || size > 4 )
         {
            data_pos = tid.offset;
         }

         temp = tell();
         seek( data_pos );

         if( tid.tag == 0x8769 ) { Exif_IFD = tid.offset; continue; }

         if( parser( *this, tid ) )
            add_tag(IFD, *this, tid);
      }

      if( Exif_IFD )
      {
         temp2 = tell();
         seek( Exif_IFD );
         parse_tags(ParseTag, pExif);
         seek( temp2 );
      }
   }
   catch(...) {}
}

////////////////////////////////////////////////////////////////////////////////


bool NEF_IFD::read_exif( FILE* in )
{
   try {
#ifndef _MSC_VER
   try { // __except
#endif

   pMakerNote = &MakerNote;
   pExif = &EXIF;
   pIFD = &IFD0;

   int marker;
   int marker_count = 0;
   unsigned segment_pos, segment_size;

   for(;; fseek( in, segment_pos + segment_size, SEEK_SET ) )
   {
      // wait for marker
      do
      {
         marker = fgetc( in );
         if (marker == EOF)
            { return false; }
      }
      while (marker == 0xFF);

      // store start of marker segment
      segment_pos = ftell( in );

      if( marker_count == 0 && marker != 0xD8 ) // SOI
         { return false; }

      ++marker_count;

      if( marker == 0xD8 ) // SOI, not size
      {
         segment_size = 0;
         continue;
      }

      if( marker == 0xD9 ) // EOI
         break;

      if( marker == 0xDA ) // SOS
         break;

      segment_size = fgetc( in )*256 + fgetc( in );
      if( feof( in ) ) { break; }

      if( marker == 0xE1 ) // APP1
      {
         TBuffer buff( segment_size );
         fread( buff, segment_size, 1, in );

         if( memcmp( buff, "Exif\0", 5) != 0 ) { continue; }

         TObjTIFF tiff( buff+6, segment_size-6, segment_pos+6);

         if( !tiff.init() )
            { continue; }

         // IFD0
         unsigned temp = tiff.get_dword();
         if( !temp ) break;

         tiff.seek( temp );
         tiff.parse_tags( ParseTag, pIFD );
         return true;
      }
   }
#ifndef _MSC_VER
   } __except(1){}
#endif
   } catch(...){}

   return false;
}
