#define STRICT
#include <windows.h>
#include <commctrl.h>
#pragma hdrstop

#include "raw2nef_setup.rc"

const char* ProgName = "RAW to NEF Command Setup";
HINSTANCE hAppInst;
HWND hAppWin;
HHOOK hHook;

bool compatible_option = false;
bool verbose_option = false;
int camera = 0;

struct TCameraItem { char option[8]; char name[32]; }
   cameras[] = {
   {"a", "AUTO"},
   {"E2100", "Nikon Coolpix 2100"},
   {"E2500", "Nikon Coolpix 2500"},
   {"E3100", "Nikon Coolpix 3100"},
   {"E3700", "Nikon Coolpix 3700"},
   {"E4300", "Nikon Coolpix 4300"},
   {"E4500", "Nikon Coolpix 4500"},
   {"E5000", "Nikon Coolpix 5000"},
   {"E5700", "Nikon Coolpix 5700"},
   {"E775", "Nikon Coolpix 775"},
   {"E885", "Nikon Coolpix 885"},
   {"E995", "Nikon Coolpix 995"},
   {"E990", "Nikon Coolpix 990"},
   {"E950", "Nikon Coolpix 950"},
   {"E900", "Nikon Coolpix 900"},
   {"C2020Z", "Olympus Camedia C2020Z"},
   {"C4100Z", "Olympus Camedia C4000Z/C4100Z"},
   {"C4040Z", "Olympus Camedia C4040Z"},
   {"C40Z",   "Olympus Camedia C40Z/D40Z"},
   {"Z2", "Konica Minolta Z2"},
};

void ShowCommandLine(HWND hDlg)
{
   char command_line[32];

   int len = 0;

   if( verbose_option )
      len += wsprintf(command_line+len, " -v");

   if( compatible_option )
      len += wsprintf(command_line+len, " -c");

   if( camera )
      len += wsprintf(command_line+len, " -%s", cameras[camera].option );

   wsprintf(command_line+len, " \"%%1\"");

   SetDlgItemText(hDlg, IDC_COMMAND_LINE, command_line);
}

static const char* GetHelpText(int id)
{
   switch(id)
   {
   case IDOK: return "Add command \"RAW to NEF\" to registry.";
   case IDCANCEL: return "Close this dialog.";
   case IDC_VERBOSE: return "Show detailed report after conversion.";
   case IDC_COMMAND_LINE: return "Command line arguments for RAW to NEF application. You can change it before press [Add] button.";
   case IDC_DELETE: return "Delete command \"RAW to NEF\" from registry.";
   case IDC_COMPATIBLE: return "If this flag is checked image has full size and can be used with any applications supporting E5700/E8700 NEF files.";
   case IDC_EXECUTABLE: return "Windows Explorer execute this application for RAW to NEF conversion.";
   case IDC_CAMERA: return "You may specify your camera or keep AUTO. You must select E990 or E885 if your camera is Coolpix 990 or Coolpix 885 correspondingly.";
   case IDC_COMMAND_NAME: return "Windows Explorer will show this command name in the context menu for the RAW files.";
   case IDC_BROWSE: return "Choise RAW to NEF convertor application.";
   case IDC_EXTENSION: return "Type file extension which you use usually for the Coolpix RAW files.";
   default: return NULL;
   }
}
static last_id = -1;
LRESULT CALLBACK GetMsgProc(int nCode, WPARAM wParam, LPARAM lParam)
{
   MSG& msg = *(MSG*)lParam;

   if( nCode >= 0 && msg.message == WM_MOUSEMOVE )
   {
      int id = GetWindowLong( msg.hwnd, GWL_ID );
      if( id != last_id )
      {
         const char* s = GetHelpText( id );
         SetDlgItemText(hAppWin, IDC_COMMENT, s ? s : "" );
         last_id = id;
      }
   }
   return (CallNextHookEx(hHook, nCode, wParam, lParam));
}

void InitDlg(HWND hDlg)
{
   hHook = SetWindowsHookEx(WH_GETMESSAGE, GetMsgProc,
        (HINSTANCE) NULL, GetCurrentThreadId());

   char FilePath[MAX_PATH];

   CheckDlgButton(hDlg, IDC_VERBOSE, verbose_option ? BST_CHECKED : BST_UNCHECKED);
   CheckDlgButton(hDlg, IDC_COMPATIBLE, compatible_option ? BST_CHECKED : BST_UNCHECKED);
   SetDlgItemText(hDlg, IDC_COMMAND_LINE, "");

   GetModuleFileName(NULL, FilePath, sizeof(FilePath));
   char* fn = strrchr(FilePath, '\\');
   if( !fn ) fn = strrchr(FilePath, '/');
   if( !fn ) fn = FilePath; else ++fn;
   strcpy( fn, "raw2nefw.exe" );

   SetDlgItemText(hDlg, IDC_EXECUTABLE, FilePath);
   SetDlgItemText(hDlg, IDC_COMMAND_NAME, "RAW to NEF");
   SetDlgItemText(hDlg, IDC_EXTENSION, ".JPG");

   int i, j=CB_ERR;
   for( int k=0; k< sizeof(cameras)/sizeof(cameras[0]); ++k)
   {
      i = (int)SendDlgItemMessage(hDlg, IDC_CAMERA, CB_ADDSTRING, 0, (LPARAM)cameras[k].name );
      if( i != CB_ERR ) SendDlgItemMessage(hDlg, IDC_CAMERA, CB_SETITEMDATA, (WPARAM)i, k );
      if( k == camera ) j = i;
   }

   if( j != CB_ERR )
      SendDlgItemMessage(hDlg, IDC_CAMERA, CB_SETCURSEL, j, 0);

   ShowCommandLine(hDlg);
}

char buffer[256];

struct TKey
{
   HKEY key;
   operator HKEY () const { return key; }
   HKEY* operator & () { return &key; }
   TKey(HKEY k=0) : key(k) {}
   ~TKey() { if( key ) RegCloseKey(key); }

   LONG SetValue(const char* name, const char* value) { return RegSetValue(key, name, REG_SZ, value, strlen(value)+1); }
   LONG SetValue(const char* name, const char* value, int size) { return RegSetValue(key, name, REG_SZ, value, size); }
};

bool UpdateRegistry(HWND hDlg, bool create_command=true)
{
   char extension[16];
   char filetype[64];
   char command_name[64];
   char exec_path[MAX_PATH];

   GetDlgItemText(hDlg, IDC_EXTENSION, extension, sizeof(extension));
   GetDlgItemText(hDlg, IDC_COMMAND_NAME, command_name, sizeof(command_name));
   GetDlgItemText(hDlg, IDC_EXECUTABLE, exec_path, sizeof(exec_path));

   TKey file_key;
   LONG err, size;

   TKey ext_key;
   err = RegOpenKey(HKEY_CLASSES_ROOT, extension, &ext_key);
   if( err != ERROR_SUCCESS )
   {
      if( err != ERROR_FILE_NOT_FOUND || !create_command )
      {
         wsprintf(buffer, "OpenKey: HKEY_CLASSES_ROOT\\%s, error=%u", extension, err);
         return false;
      }

      err = RegCreateKey(HKEY_CLASSES_ROOT, extension, &file_key);
      if( err != ERROR_SUCCESS )
      {
         wsprintf(buffer, "CreateKey: HKEY_CLASSES_ROOT\\%s, error=%u", extension, err);
         return false;
      }
      file_key.SetValue( NULL, "Coolpix Raw File" );
   }
   else
   {
      size = sizeof(filetype);
      err = RegQueryValue(ext_key, "", filetype, &size);
      if( err != ERROR_SUCCESS )
      {
         wsprintf(buffer, "QueryValue: HKEY_CLASSES_ROOT\\%s, Default, error=%u", extension, err);
         return false;
      }

      err = RegOpenKey(HKEY_CLASSES_ROOT, filetype, &file_key);
      if( err != ERROR_SUCCESS )
      {
         wsprintf(buffer, "Open: HKEY_CLASSES_ROOT\\%s, error=%u", filetype, err);
         return false;
      }
   }

   if( !create_command )
   {
      wsprintf(buffer, "shell\\%s\\command", command_name);

      err = RegDeleteKey(file_key, buffer);
      if( err != ERROR_SUCCESS )
      {
         wsprintf(buffer, "DeleteKey: HKEY_CLASSES_ROOT\\%s\\shell\\%s\\command, error=%u", filetype, command_name, err);
         return false;
      }

      wsprintf(buffer, "shell\\%s", command_name);

      err = RegDeleteKey(file_key, buffer);
      if( err != ERROR_SUCCESS )
      {
         wsprintf(buffer, "DeleteKey: HKEY_CLASSES_ROOT\\%s\\shell\\%s, error=%u", filetype, command_name, err);
         return false;
      }

      wsprintf(buffer, "DeleteKey: HKEY_CLASSES_ROOT\\%s\\shell\\%s", filetype, command_name);
      return true;
   }
   else
   {
      TKey command_key;
      wsprintf(buffer, "shell\\%s\\command", command_name);

      err = RegCreateKey(file_key, buffer, &command_key);
      if( err != ERROR_SUCCESS )
      {
         wsprintf(buffer, "CreateKey: HKEY_CLASSES_ROOT\\%s\\shell\\%s\\command, error=%u", filetype, command_name, err);
         return false;
      }

      size = wsprintf(buffer, "\"%s\"", exec_path);
      
      GetDlgItemText(hDlg, IDC_COMMAND_LINE, &buffer[size], sizeof(buffer)-size);

      err = command_key.SetValue( NULL, buffer);
      if( err != ERROR_SUCCESS )
      {
         wsprintf(buffer, "SetValue: HKEY_CLASSES_ROOT\\%s\\shell\\%s\\command, error=%u", filetype, command_name, err);
         return false;
      }
   }

   wsprintf(buffer, "SetValue: HKEY_CLASSES_ROOT\\%s\\shell\\%s\\command", filetype, command_name);
   return true;
}

void Browse(HWND hDlg)
{
   char bufFileName[MAX_PATH]; bufFileName[0] = '\0';

   OPENFILENAME ofn;
   memset( &ofn, 0, sizeof(ofn) );

   ofn.lStructSize = sizeof(ofn);
   ofn.hwndOwner = hDlg;
   ofn.hInstance = hAppInst;
   ofn.lpstrFilter = "Raw2Nef*.exe\0raw2nef*.exe\0"
      "Exe Files (*.exe)\0*.exe\0"
      "All Files (*.*)\0*.*\0";
   ofn.lpstrFile = bufFileName;
   ofn.nMaxFile = sizeof(bufFileName);
   ofn.Flags = OFN_EXPLORER|OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;

   if( !GetOpenFileName( &ofn ) )
      return;

   SetDlgItemText(hDlg, IDC_EXECUTABLE, bufFileName );
}

BOOL Command(HWND hDlg, int NotifyCode, int id, HWND hCtrl)
{
   NotifyCode;
   hCtrl;

   switch(id)
   {
   case IDCANCEL:
      EndDialog(hDlg, 0);
      return TRUE;

   case IDOK:
      // UpdateRegistry
      if( IDYES == MessageBox(hDlg, "Add command to registry?", ProgName, MB_ICONQUESTION|MB_YESNO) )
      {
         if( UpdateRegistry(hDlg, true) )
         {
            strcat(buffer, "\r\nCommand has added to registry");
            MessageBox(hDlg, buffer, ProgName, MB_ICONINFORMATION|MB_OK);
         }
         else
         {
            strcat(buffer, "\r\nCannot setup RAW to NEF command");
            MessageBox(hDlg, buffer, ProgName, MB_ICONSTOP|MB_OK);
         }
      }
      return TRUE;

   case IDC_DELETE:
      if( IDYES == MessageBox(hDlg, "Delete command from registry?", ProgName, MB_ICONQUESTION|MB_YESNO) )
      {
         if( UpdateRegistry(hDlg, false) )
         {
            strcat(buffer, "\r\nCommand has deleted from registry");
            MessageBox(hDlg, buffer, ProgName, MB_ICONINFORMATION|MB_OK);
         }
         else
         {
            strcat(buffer, "\r\nCannot delete RAW to NEF command");
            MessageBox(hDlg, buffer, ProgName, MB_ICONSTOP|MB_OK);
         }
      }
      return TRUE;

   case IDC_CAMERA:
      if( NotifyCode == CBN_SELCHANGE )
      {
         int i = SendDlgItemMessage(hDlg, IDC_CAMERA, CB_GETCURSEL, 0, 0);
         if( i != CB_ERR ) camera = (unsigned)SendDlgItemMessage(hDlg, IDC_CAMERA, CB_GETITEMDATA, (WPARAM)i, 0 );
         goto show_command_line;
      }
      return TRUE;

   case IDC_VERBOSE:
      verbose_option = !verbose_option;
      goto show_command_line;

   case IDC_COMPATIBLE:
      compatible_option = !compatible_option;
      
   show_command_line:
      ShowCommandLine(hDlg);
      return TRUE;

   case IDC_BROWSE:
      Browse(hDlg);
      return TRUE;

   default:
      return FALSE;
   }
}

BOOL CALLBACK MainDlg( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
   if( uMsg == WM_INITDIALOG )
   {
      if( !hAppWin )
         hAppWin = hDlg;

      InitDlg(hDlg);
      return TRUE;
   }

   if( uMsg == WM_COMMAND )
   {
      return Command(hDlg, HIWORD(wParam), LOWORD(wParam), (HWND) lParam);
   }

   if( uMsg == WM_CLOSE )
   {
      EndDialog(hDlg, 0);
      return TRUE;
   }

   if( uMsg == WM_DESTROY )
   {
      if( hDlg == hAppWin )
         hAppWin = NULL;

      if( hHook )
      {
          UnhookWindowsHookEx( hHook );
          hHook = NULL;
      }
      return FALSE;
   }
   return FALSE;
}

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE, LPSTR, int )
{
   hAppInst = hInstance;

   InitCommonControls();

   DialogBox( hAppInst, MAKEINTRESOURCE(IDD_DIALOG1), NULL, MainDlg);

   return 0;
}

